///prop2harm, subfunction in Propose. MEX implementation utilizing OpenMP
//This function governs the calculation of nonlinear wave propagation in Propose.
///Master student Haakon Seljaasen fall 2013.
#include "math.h"
#include "mex.h"
#include "matrix.h"
#include "complex.h"
using namespace std;

#define ompWORKFLOW dynamic //Open MP schedule() setting -affects all calls

#ifdef _OPENMP
    #include "headerFiles/omp.h"
#else
    #define omp_get_thread_num() 0
#endif


//subfunctions:
double complex nodePointCalculation(const mxArray * P11, const mxArray * K,int fxStart,int fxStop,int fyStart,int fyStop,int fStart,int fStop,double z1,double z2,double fx,double fy,double f,double a,double b,double c,double complex Cfactor,double infinitesimalProduct,int m,int n,int l,int NfHalf);
double complex katet(double fx, double fy, double f, double * a, double * b, double c);
mxArray * katet(const double* fx, int Nfx1, const double* fy, int Nfy1, const double* f, int fLen, const double* a, const double* b, const double * c);
double mean (const double array[], const double length);
double complex csinc (double complex x);
mwIndex singleIndex(int i, int j, int k, int sizeI, int sizeJ, bool baseOfIndex);
void tripleIndex(mwIndex singleIndex, int indexArray[3], int sizeI, int sizeJ, bool baseOfReturnIndex);
void utlizeSymmetry(double complex P221, double * outputRe, double * outputIm,int index[3], int dims[3],int Nx2, int Ny2, bool baseOfInputIndex);

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]){
    /*The function prop3harm is computing the 2. harmonic nonlinear propagation from plane P11 to P22, using the
     *quasi-linear approximation. Input parameters are (P11,fx1,fy2,f1,z1,z2,a,b), output parameters are (P22,fx2,fy2,f2).
     * - P11 is the Fouriertransform of the field in pos. z1, with freq. axes fx1,fy1,f1
     * - P22 is the Fouriertransform of the 2. harm field in pos. z2, with freq. axes fx1,fy1,f1.*/
//Tested 29.12.13 and working correct, Haakon Seljaasen.
//Validation of inputs:
    if (nlhs!=4){
        mexErrMsgIdAndTxt("Toolbox:prop2harm:nlhs","exactly 4 output variables is allowed (P22,fx2,fy2,f2)");
    };
    if (nrhs!=8){
        mexErrMsgIdAndTxt("Toolbox:prop2harm:nrhs","exactly 8 inputs are needed: (P11,fx1,fy2,f1,z1,z2,a,b)");
    };
    //Definition of pi:
    long double pi = 3.1415926535897932384626433832795028841971693993751;  //(double)(*PI);
    //Calculation method:
    bool useSymmetry;//Assigned from pSettings in Propose
    //Declaration of variables
    const mxArray * P11=prhs[0];
    const mxArray * Z1=prhs[4];
    const mxArray * Z2=prhs[5];
    const double * fx1=mxGetPr(prhs[1]);
    const double * fy1=mxGetPr(prhs[2]);
    const double * f1=mxGetPr(prhs[3]);
    mxArray * Fx1=mxDuplicateArray(prhs[1]);
    mxArray * Fy1=mxDuplicateArray(prhs[2]);
    mxArray * Ff1=mxDuplicateArray(prhs[3]);
    double * fx2;
    double * fy2;
    double * f2;
    double * z1=mxGetPr(prhs[4]);
    double * z2=mxGetPr(prhs[5]);
    double * a=mxGetPr(prhs[6]);
    double * b=mxGetPr(prhs[7]);
    mxArray * A=mxDuplicateArray(prhs[6]);
    mxArray * B=mxDuplicateArray(prhs[7]);
    mxArray * P22;
    mxArray * Fx;
    mxArray * Fy;
    
    double * P22Re=NULL;
    double * P22Im=NULL;
    //Various complex variables:
    double complex Kf;
    double complex P221;
    double complex ImgPropFactor;
    //Various framework variables:
    double temp=0;
    int Nfx1, Nfy1, Nf1, Nx, Ny, Nf, Nx2, Ny2, Nf2, NfHalf, Nl, Nm, Nn;
    int fxStart,fxStop,fyStart,fyStop,fStart,fStop;
    double dfx, dfy, df, f0, limit, f, fy, fx;
    mxArray * K;
    mwIndex sizeOfMatrix;
    int oneBasedIndex[3];
    double kappa, betaN, c;//Acoustical properties
    //Extracting setting values from the global workspace, pSettings.
    //-----------------------------------------------------------
    mxArray *pSettings=mexGetVariable("global", "pSettings"); // get the struct
    if (pSettings==NULL){//error testing
        //mexPrintf("\nError in prop2harm.cpp: Struct pSettings not found\n");
        mexErrMsgIdAndTxt("Toolbox:prop2harm:pSettings","Matlab variable pSettings not found");
    };
    //-------------------------useSymmetry?-------------------------------
    mxArray *Symmetry = mxGetField(pSettings,0,"useSymmetry"); // get the field
    if (Symmetry==NULL){//error testing
        mexErrMsgIdAndTxt("Toolbox:prop2harm:Symmetry","Struct value pSettings.useSymmetry not found");
    };
    if ((*mxGetPr(Symmetry)==0.0)||(*mxGetPr(Symmetry)==1)){//error testing
        useSymmetry=(bool)(*mxGetPr(Symmetry));
    }else{
        mexErrMsgIdAndTxt("Toolbox:prop2harm:Symmetry","Struct value pSettings.useSymmetry contains non-boolean value");
    };
    
    mxArray *Kappa = mxGetField(pSettings,0,"kappa"); // get the field
    if (Kappa==NULL){//error testing
        mexErrMsgIdAndTxt("Toolbox:prop2harm:Kappa","Struct value pSettings.Kappa not found");
    };
    kappa=(*mxGetPr(Kappa));
    mxArray *BetaN = mxGetField(pSettings,0,"betaN"); // get the field
    if (BetaN==NULL){//error testing
        mexErrMsgIdAndTxt("Toolbox:prop2harm:BetaN","Struct value pSettings.BetaN not found");
    };
    betaN=(*mxGetPr(BetaN));
    mxArray *SpeedOfSound = mxGetField(pSettings,0,"c"); // get the field
    if (SpeedOfSound==NULL){//error testing
        mexErrMsgIdAndTxt("Toolbox:prop2harm:SpeedOfSound","Struct value pSettings.c not found");
    };
    c=*(mxGetPr(SpeedOfSound));
    ImgPropFactor=0-(kappa*betaN*pow(pi,2)/pow(c,4))*I;
    //Finding properties of input Matrix.
    Nfx1=mxGetN(prhs[1]);
    Nfy1=mxGetN(prhs[2]);
    Nf1=mxGetN(prhs[3]);
    mwSize ndim=3; //3D Matrix output
    //Extracting dimmensions from the input Matrix:
    const mwSize * P11Dims=mxGetDimensions(P11);
    Nx=P11Dims[0];
    Ny=P11Dims[1];
    Nf=P11Dims[2];
    mwIndex tempIndex[3];
    //Allocating mxArrays and pointers for output:
    tempIndex[0]=1;     //Needed to make the output Arrays column major.
    tempIndex[1]=2*Nx;  //using this temp. variable to satisfy input requirements in mxCreateNumericArray.
    plhs[1]=mxCreateNumericArray(2, tempIndex, mxDOUBLE_CLASS, mxREAL);//Allocating output mxArray
    fx2=mxGetPr(plhs[1]);//Making double pointer to output, for use inside prop2harm.
    tempIndex[1]=2*Ny;
    plhs[2]=mxCreateNumericArray(2, tempIndex, mxDOUBLE_CLASS, mxREAL);
    fy2=mxGetPr(plhs[2]);//Making double pointer to output, for use inside prop2harm.
    tempIndex[1]=Nf;
    plhs[3]=mxCreateNumericArray(2, tempIndex, mxDOUBLE_CLASS, mxREAL);
    f2=mxGetPr(plhs[3]);//Making double pointer to output, for use inside prop2harm.
    dfx=fx1[1]-fx1[0];
    dfy=fy1[1]-fy1[0];
    df=f1[1]-f1[0];
    double infinitesimalProduct=dfx*dfy*df;
    //Frequency axes for 2. harmonic:
    f0=mean(f1,Nf1);
    Nf2=Nf1;
    for(int i=0;i<Nf1;i++){
        f2[i]=f1[i]+f0;
    };
//Initializing counters for determination of array length Nx2 and Ny2
    Nx2=0;
    temp=2*fx1[0];
    limit=2*fx1[Nfx1-1]+dfx;
    while(temp<limit){
        fx2[Nx2]=2*fx1[0]+dfx*(Nx2);
        temp=fx2[Nx2];
        //mexPrintf("\nfx2=%f\n",temp);
        Nx2++;
    };
    Ny2=0;
    temp=2*fy1[0];
    limit=2*fy1[Nfy1-1]+dfy;
    while(temp<limit){
        fy2[Ny2]=2*fy1[0]+dfy*(Ny2);
        temp=fy2[Ny2];
        //mexPrintf("\nfy2=%f\n",temp);
        Ny2++;
    };
    /*If one wish to include dispersion this can be included here,
     * this function may be called from matlab: cw=dispersion(f2,2*f0,a,b)*/
    //Creating output Matrix:
    mwSize P22Dims[3];
    P22Dims[0]=Nx2;
    P22Dims[1]=Ny2;
    P22Dims[2]=Nf2;
    plhs[0]=mxCreateNumericArray(ndim, P22Dims, mxDOUBLE_CLASS, mxCOMPLEX);
    P22=plhs[0];
    sizeOfMatrix=mxGetNumberOfElements(P22);
    //Directing return Matrix pointers;
    P22Re=mxGetPr(P22);
    P22Im=mxGetPi(P22);
    //Making Matrix/table with K values for use in the nodePointCalculation function:
    K=katet(fx1, Nfx1, fy1, Nfy1, f1 , Nf1, a, b, &c); //Calculating katet() values for every combination of fx, fy and f...
    //Boundaries for the tripleIndecies
    NfHalf=floor(Nf/2.0);
    Nl=Nf;
    if(useSymmetry){
        Nm=round(Nx2/2.0);
        Nn=round(Ny2/2.0);
    }else{
        Nm=Nx2;
        Nn=Ny2;
    };
    
    //Main calculations:
    if ( ! useSymmetry ){
#pragma omp parallel for private(fy,fx,f,fyStart,fyStop,fxStart,fxStop,fStart,fStop,P221,oneBasedIndex) shared(P22Dims,P11,K,z1,z2,a,b,c,ImgPropFactor,infinitesimalProduct,NfHalf,sizeOfMatrix) schedule(ompWORKFLOW) //Calling openMP
        for (mwIndex t=0; t<sizeOfMatrix; t++){
            
            tripleIndex(t, oneBasedIndex, P22Dims[0], P22Dims[1], 1);
//The oneBasedIndex values corresponds to the following indecies used in the original MATLAB code.
//      m=oneBasedIndex[0]; //x
//      n=oneBasedIndex[1]; //y
//      l=oneBasedIndex[2]; //f
            fx=fx2[oneBasedIndex[0]-1];
            fxStart=fmax(1,1+oneBasedIndex[0]-Nx);
            fxStop=fmin(Nx-1,oneBasedIndex[0]);
            
            fy=fy2[oneBasedIndex[1]-1];
            fyStart=fmax(1,1+oneBasedIndex[1]-Ny);
            fyStop=fmin(Ny-1,oneBasedIndex[1]);
            
            f=f2[oneBasedIndex[2]-1];
            fStart=fmax(1,1+oneBasedIndex[2]-NfHalf);
            fStop=fmin(Nf,oneBasedIndex[2]-NfHalf+Nf-1);
            if  (!((fStart>fStop)||(fyStart>fyStop)||(fxStart>fxStop))){
                P221=nodePointCalculation(P11,K,fxStart,fxStop,fyStart,fyStop,fStart,fStop,*z1,*z2,fx,fy,f,*a,*b,c,ImgPropFactor,infinitesimalProduct,oneBasedIndex[0],oneBasedIndex[1],oneBasedIndex[2],NfHalf);
            }else{
                P221=0+0*I;
            };
            P22Re[t]=creal(P221);
            P22Im[t]=cimag(P221);
        }; // end of for (t=0 to sizeOfMatrix)
    }else{
//----------------------if symmetry is enabled ---------------------------
#pragma omp parallel for private(fy,fx,f,fyStart,fyStop,fxStart,fxStop,fStart,fStop,P221,oneBasedIndex) shared(P22Dims,P11,K,z1,z2,a,b,c,ImgPropFactor,infinitesimalProduct,NfHalf) schedule(ompWORKFLOW) //Calling openMP
        for (int l=1; l<=Nl; l++){
            f=f2[l-1];
            fStart=fmax(1,1+l-NfHalf);
            fStop=fmin(Nf,l-NfHalf+Nf-1);
            for (int m=1;m<=Nm;m++){
                fx=fx2[m-1];
                fxStart=fmax(1,1+m-Nx);
                fxStop=fmin(Nx-1,m);
                for (int n=1;n<=Nn;n++){
                    fy=fy2[n-1];
                    fyStart=fmax(1,1+n-Ny);
                    fyStop=fmin(Ny-1,n);
                    if  (!((fStart>fStop)||(fyStart>fyStop)||(fxStart>fxStop))){
                        P221=nodePointCalculation(P11,K,fxStart,fxStop,fyStart,fyStop,fStart,fStop,*z1,*z2,fx,fy,f,*a,*b,c,ImgPropFactor,infinitesimalProduct,m,n,l,NfHalf);
                    }else{
                        P221=0+0*I;
                    };
                    oneBasedIndex[0]=m;
                    oneBasedIndex[1]=n;
                    oneBasedIndex[2]=l;
                    utlizeSymmetry( P221,P22Re,P22Im,oneBasedIndex,P22Dims,Nx2,Ny2,1 );
                };//for m=0 to Nn
            }; //for l=0  to Nm
        }; //for l=0 to Nl
    };//end of if-else statement regarding symmetry
    
    return;
};


double complex nodePointCalculation(const mxArray * P11, const mxArray * K,int fxStart,int fxStop,int fyStart,int fyStop,int fStart,int fStop,double z1,double z2,double fx,double fy,double f,double a,double b,double c,double complex Cfactor,double infinitesimalProduct,int m,int n,int l,int NfHalf){
    /*The function nodePointCalculation is used for computing the 2. harmonic nonlinear soundfield of a node in P22 of prop2harm, using the
     *quasi-linear approximation. Input parameters are (P11,K,kf,fxStart,fxStop,fyStart,fyStop,fStart,fStop,z1,z2,fx,fy,f,a,b,c,Cfactor,dfx,dfy,df,l,m,n,NfHalf) */
//Tested 29.12.13 and working correct, HS.
    double pi = 3.1415926535897932384626433832795028841971693993751;
    //Output:
    double complex Node=0+0*I;
    //Input
    double * P11Re=mxGetPr(P11);
    double * P11Im=NULL;
    bool complexP11=mxIsComplex(P11);
    if (complexP11){
        P11Im=mxGetPi(P11);
    }else{
        mexErrMsgIdAndTxt("Toolbox:prop2harm:nodePointCalculation","Non-complex P11 input given, needs to be taken into account in the prop2harm.cpp code before enabeling.");
    };
    bool complexK=mxIsComplex(K);
    double * K_Real=mxGetPr(K);
    double * K_Imag=NULL;
    if (complexK){
        K_Imag=mxGetPi(K);//Taken into account later in the code with the complexK-variable
    };
    double complex kf=katet(fx, fy, f, &a, &b, c);
    const mwSize * Kdim=mxGetDimensions(K);
    const mwSize * P11dim=mxGetDimensions(P11);
    //Working variables
    double complex gamma;
    double complex K1;
    double complex K2;
    double complex P11A;
    double complex P11B;
    double Z1minusZ2_Half=((z1)-(z2)/2);//for use in Hp
    double MinusZ2div2pi=-(z2)/(2*pi);//for use in Hp
    mwIndex singleIndex1;
    mwIndex singleIndex2;
    mwIndex Ix2, Iy2, If2;
    double complex Hp;
    //Calculations
    for (int h=fxStart; h<=fxStop; h++){
        Ix2=1+m-h;
        for(int i=fyStart; i<=fyStop; i++){
            Iy2=1+n-i;
            for(int j=fStart; j<=fStop; j++){
                If2=l+1+NfHalf-j;
                //Calculating single index
                singleIndex1=singleIndex(h, i, j, Kdim[0], Kdim[1],1);
                singleIndex2=singleIndex(Ix2, Iy2, If2, Kdim[0], Kdim[1],1);
                if(complexK){
                    K1=K_Real[singleIndex1]+K_Imag[singleIndex1]*I;
                    K2=K_Real[singleIndex2]+K_Imag[singleIndex2]*I;
                }else{
                    K1=K_Real[singleIndex1]+0*I;
                    K2=K_Real[singleIndex2]+0*I;
                };
                gamma=K1+K2-kf;
                Hp=cexp(gamma*I*Z1minusZ2_Half)*csinc(gamma*MinusZ2div2pi);
                //Recalculating single index
                singleIndex1=singleIndex(h, i, j, P11dim[0], P11dim[1],1);
                singleIndex2=singleIndex(Ix2, Iy2, If2, P11dim[0], P11dim[1],1);
                P11A=P11Re[singleIndex1]+P11Im[singleIndex1]*I;
                P11B=P11Re[singleIndex2]+P11Im[singleIndex2]*I;
                Node=Node+(P11A*P11B*Hp);
            };
        };
    };
    Node=Node*cexp( ( -kf*I )*( z2-z1 ) )*z2*pow( f,2 )*infinitesimalProduct;
    Node=Node*( Cfactor / ( kf+2*pi*f/c ) );
    return Node;
};

double complex katet(double fx, double fy, double f, double * a, double * b, double c){
    /*The function katet is ment for computing sqrt(f^2-f_x^2-f_y^2). Input is
     * (fx[],fy[],f[],a,b,c) Output is double a scalar that is negative if the answer is complex. a and b are not used unless we want to include attenuation in the formula.*/
//Definition of pi:
    double pi = 3.1415926535897932384626433832795028841971693993751;
    int int_b=*b;
    double complex K;
    double arg=pow(f,2)-pow(fx,2)-pow(fy,2);
    K=( csqrt(arg)-f )*( (2*pi)/(c) );
    //Subtracting attenuation:
    K=K-(*a)*pow(f/pow(10.0,6),int_b)*I;
    return K;
};

//katet.cpp - Haakon Seljaasen//
mxArray * katet(const double* fx, int Nfx1, const double* fy, int Nfy1, const double* f, int fLen, const double* a, const double* b, const double * c){
    /*The function katet is ment for computing sqrt(f^2-f_x^2-f_y^2). Input is
     * (fx[],Nfx,fy[],Nfy,f[],Nf,a,b,c)*/
//Definition of pi:
    double pi = 3.1415926535897932384626433832795028841971693993751;
//Declaration of variables
    mxArray * output;
    double * sqrtRe=NULL;
    double * sqrtIm=NULL;
    int int_b=*b;
    double proportionalityFactor=(2*pi)/(*c);
    double arg;
    double complex K;
//Declaring dimmensions on the output array.
    mwSize ndim=3;
    mwSize dims[3];
//Creating output Matrix
//Set dimmensions:
    dims[0]=Nfx1;
    dims[1]=Nfy1;
    dims[2]=fLen;
    mxClassID classid=mxDOUBLE_CLASS;
    mxComplexity ComplexFlag=mxCOMPLEX;
    mwIndex It;
    if((Nfx1==1)&&(Nfy1==1)&&(fLen)){
        output=mxCreateDoubleMatrix(1, 1, mxCOMPLEX);
    }else{
        output=mxCreateNumericArray(ndim, dims, classid,ComplexFlag);
        sqrtIm=mxGetPi(output);
    };
    sqrtRe=mxGetPr(output);
//    #pragma omp parallel for private(arg,root,K, attenuation,It) schedule(ompWORKFLOW) //Usually too small matrixes to make any positive difference
    //The function is only called once.
    for(int i=0;i<fLen;i++){
        for(int j=0;j<Nfx1;j++){
            for(int k=0;k<Nfy1;k++){
                /* According to original matlab souce code here we make a change
                 * to make sure that the imaginary part of K is negative, so that the solution
                 * does not diverge as z approaches infinity. */
                arg=pow(f[i],2)-pow(fx[j],2)-pow(fy[k],2);
                if
                K=(csqrt(arg)-f[i])*proportionalityFactor;
                //Accounting for power law attenuation:
                K=K-((*a)*pow(f[i]/pow(10.0,6),int_b))*I;
                //Shipping out results:
                It=singleIndex(j, k, i, dims[0], dims[1],0);
                sqrtRe[It]=creal(K);
                sqrtIm[It]=cimag(K);
            };
        };
    };
    return output;
};

double complex csinc (double complex x){
    //Definition of pi:
    double pi = 3.1415926535897932384626433832795028841971693993751;
    double complex temp;
    if((creal(x)==0.0)&&(cimag(x)==0.0)){
        return 1+0*I;
    }else{
        x=x*pi;//Normalized sinc uses x times pi
        return csin(x)/x;
    };
};

double mean (const double array[], const double length){
    double sum = 0;
    for (int i = 0; i < length; i++){
        sum =sum+ array [i];
    }
    return sum/length;
};

mwIndex singleIndex(int i, int j, int k, int sizeI, int sizeJ, bool baseOfIndex){
    /*Creates a 0-based single subscript from the 1-based matlab indexi of a 3D matrix
     * if baseOfIndex==1, and a 0 base index of 0 based indices if baseOfIndex==0 indices of a 3D matrix.
     * i, j, k are indices wished to access, I and J denotes the first two dimmensions of the 3D Matrix*/
    if(baseOfIndex){
        return (i-1)+(j-1)*sizeI+(k-1)*sizeI*sizeJ;
    }else{
        return (i)+(j)*sizeI+(k)*sizeI*sizeJ;
    };
};

void tripleIndex(mwIndex singleIndex, int indexArray[3], int sizeI, int sizeJ, bool baseOfReturnIndex){
    /*The function trippleindex calculates three zero based (baseOfReturnIndex=0) or 1-based (baseOfReturnIndex=1) indices from a C++ MEX 0-based singleIndex
     *used to access a mxArray in a MEX function, and the 1st and 2nd array lengths.*/
    //Basis equation is: singleIndex= i + sizeI*j + sizeI*sizeJ*k;
    int IxJ = sizeI * sizeJ;
    int rest = singleIndex % IxJ; //i.e. rest=sizeI*j + i;
    indexArray[0]=( rest % sizeI ); //This is the final i value (1st index)
    indexArray[1]=( ( rest - indexArray[0] ) / sizeI ); //This is the final j value (2nd index)
    indexArray[2]=( ( singleIndex - indexArray[0] - indexArray[1] * sizeI ) / IxJ ); //This is the final k value (3rd index)
    if (baseOfReturnIndex){
        indexArray[0]++;
        indexArray[1]++;
        indexArray[2]++;
    }; //Modification of output if 1-based indexing is specified (defualt return value is 0-based)
    return;
};

void utlizeSymmetry(double complex P221, double outputRe[], double outputIm[],int index[3], int dims[3],int Nx2, int Ny2, bool baseOfInputIndex){
    ///This function is used to distribute a computed value in prop2harm to all 4 destinations in the output array where it applies according to symmetry properties.
    mwIndex tempIt;
    tempIt=singleIndex(index[0], index[1], index[2], dims[0], dims[1], baseOfInputIndex);
    outputRe[tempIt]=creal(P221);
    outputIm[tempIt]=cimag(P221);
    
    tempIt=singleIndex(Nx2-index[0]+1, index[1], index[2], dims[0], dims[1], baseOfInputIndex);
    outputRe[tempIt]=creal(P221);
    outputIm[tempIt]=cimag(P221);
    
    tempIt=singleIndex(Nx2-index[0]+1, Ny2-index[1]+1, index[2], dims[0], dims[1], baseOfInputIndex);
    outputRe[tempIt]=creal(P221);
    outputIm[tempIt]=cimag(P221);
    
    tempIt=singleIndex(index[0], Ny2-index[1]+1, index[2], dims[0], dims[1], baseOfInputIndex);
    outputRe[tempIt]=creal(P221);
    outputIm[tempIt]=cimag(P221);
    return;
};

